var ERR = require("./../server/server_errs"), fs = require("fs"), os = require("os"), path = require("path"), util = require("util"), browsers = require("./browsers"), ports = require("./ports"), validator = require("./validator");

var CFG_FILE_NAME = "config.json";

exports.CFG_FILE_DIR = path.join(__dirname, "../../");

exports.CFG_FILE_PATH = path.join(exports.CFG_FILE_DIR, CFG_FILE_NAME);

var DEFAULT_TESTS_DIR = path.join(__dirname, "../../my_tests");

var Config = exports.Config = function() {
    this.currentCfg = {};
    this.initedControlPanelPort = null;
    this.initedFromObj = false;
    this.initedServicePort = null;
};

Config.prototype.initFromObj = function(cfgObj) {
    this.initedFromObj = true;
    var errs = validator.validate(cfgObj, true);
    this.currentCfg = this._cloneCfg(cfgObj);
    this.currentCfg.hostname = this._getHostname(this.currentCfg.hostname);
    this.initedServicePort = this.currentCfg.servicePort;
    this.initedControlPanelPort = this.currentCfg.controlPanelPort;
    this.currentCfg.controlPanelHost = [ this.currentCfg.hostname, ":", this.currentCfg.controlPanelPort ].join("");
    return {
        cfg: this.currentCfg,
        errs: errs.length ? errs : null
    };
};

Config.prototype.initFromFile = function(callback) {
    var config = this;
    this._readCfgFile(function(readedCfg, readErr) {
        config._prepareUserCfg(readedCfg || {}, function(validCfg, validErrs) {
            config.currentCfg = validCfg;
            config.getConfig(function(cfg) {
                config.initedServicePort = cfg.servicePort;
                config.initedControlPanelPort = cfg.controlPanelPort;
                var readed = !!readedCfg;
                callback(cfg, readed ? readErr || validErrs : null);
            });
        });
    });
};

Config.prototype.getConfig = function(callback) {
    var config = this, result = {}, currentCfg = this.currentCfg || {};
    result.testsDir = this._getTestsDir(currentCfg.testsDir);
    result.hostname = this._getHostname(currentCfg.hostname);
    this._getPorts(function(controlPanelPort, servicePort) {
        result.controlPanelPort = controlPanelPort;
        result.servicePort = servicePort;
        result.controlPanelHost = [ result.hostname, ":", result.controlPanelPort ].join("");
        config._getBrowsers(function(browsers) {
            result.browsers = browsers;
            callback(result);
        });
    });
};

Config.prototype.setTestsDir = function(testsDir, callback) {
    if (!this.initedFromObj) {
        var fileCfg = null;
        if (!fs.existsSync(exports.CFG_FILE_PATH)) {
            fs.writeFileSync(exports.CFG_FILE_PATH, "{}");
        }
        try {
            fileCfg = JSON.parse(fs.readFileSync(exports.CFG_FILE_PATH).toString());
        } catch (err) {
            callback(err);
            return;
        }
        fileCfg.testsDir = this.currentCfg.testsDir = testsDir;
        fs.writeFile(exports.CFG_FILE_PATH, JSON.stringify(fileCfg, null, 5), callback);
    } else {
        this.currentCfg.testsDir = testsDir;
        callback();
    }
};

Config.prototype._cloneCfg = function(cfg) {
    var clone = {
        controlPanelPort: cfg.controlPanelPort,
        servicePort: cfg.servicePort,
        testsDir: cfg.testsDir,
        hostname: cfg.hostname,
        browsers: {}
    };
    for (var brName in cfg.browsers) {
        clone.browsers[brName] = {
            path: cfg.browsers[brName].path,
            cmd: cfg.browsers[brName].cmd,
            icon: cfg.browsers[brName].icon
        };
    }
    return clone;
};

Config.prototype._getTestsDir = function(cfgTestsDir) {
    if (!cfgTestsDir) {
        if (!fs.existsSync(DEFAULT_TESTS_DIR)) {
            fs.mkdirSync(DEFAULT_TESTS_DIR);
        }
        return DEFAULT_TESTS_DIR;
    }
    return cfgTestsDir;
};

Config.prototype._getHostname = function(cfgHostname) {
    return cfgHostname || os.hostname() || "127.0.0.1";
};

Config.prototype._prepareUserCfg = function(cfgObj, callback) {
    var errs = validator.validate(cfgObj);
    for (var i = errs.length - 1; i >= 0; i--) {
        delete cfgObj[errs[i].property];
    }
    callback(cfgObj, errs.length ? errs : null);
};

Config.prototype._readCfgFile = function(callback) {
    var cfgData = null, cfg = null;
    if (!fs.existsSync(exports.CFG_FILE_PATH)) {
        callback(null);
        return;
    }
    try {
        cfgData = fs.readFileSync(exports.CFG_FILE_PATH).toString();
    } catch (err) {
        callback(this.currentCfg, err);
        return;
    }
    try {
        cfg = JSON.parse(cfgData);
    } catch (parserErr) {
        callback(null, {
            code: ERR.CONFIG_PARSING_FAILED,
            parserErr: parserErr
        });
        return;
    }
    callback(cfg);
};

Config.prototype._getPorts = function(callback) {
    var config = this;
    if (!this.currentCfg || !this.currentCfg.controlPanelPort || !this.currentCfg.servicePort) {
        ports.getFreePorts(function(ports) {
            var controlPanelPort = config.currentCfg && config.currentCfg.controlPanelPort || config.initedControlPanelPort || ports.controlPanelPort, servicePort = config.currentCfg && config.currentCfg.servicePort || config.initedServicePort || ports.servicePort;
            callback(controlPanelPort, servicePort);
        });
    } else {
        callback(this.currentCfg.controlPanelPort, this.currentCfg.servicePort);
    }
};

Config.prototype._getBrowsers = function(callback) {
    if (this.currentCfg && this.currentCfg.browsers) {
        callback(this.currentCfg.browsers);
    } else {
        browsers.getBrowsers(callback);
    }
};