var fs = require("fs"), util = require("util"), javascriptParser = require("uglify-js").parser, ErrCodes = require("./err_codes");

exports.construct = function(fileName, callback) {
    fs.readFile(fileName, function(readErr, data) {
        if (readErr) {
            callback({
                code: ErrCodes.READ_FILE_FAILED,
                filename: fileName
            });
            return;
        }
        try {
            var ast = javascriptParser.parse(data.toString().trim(), false, true);
        } catch (parseErr) {
            callback({
                code: ErrCodes.JAVASCRIPT_PARSING_FAILED,
                filename: fileName,
                parserErr: parseErr
            });
            return;
        }
        callback(null, ast);
    });
};

var getEntryName = exports.getEntryName = function(entry) {
    return typeof entry[0] === "object" ? entry[0].name : entry[0];
};

exports.isPathMatch = function(expectedPath, actualPath, ensureLast) {
    if (expectedPath.length !== actualPath.length - 1) {
        return false;
    }
    if (ensureLast && !actualPath[1][0].last) {
        return false;
    }
    for (var i = 0; i < expectedPath.length; i++) {
        var hasFlag = util.isArray(expectedPath[i]), name = hasFlag ? expectedPath[i][0] : expectedPath[i];
        if (getEntryName(actualPath[i]) !== name || hasFlag && actualPath[i][1] !== expectedPath[i][1]) {
            return false;
        }
    }
    return true;
};

exports.getAncestorByName = function(name, currentAstPath) {
    if (currentAstPath.length > 1) {
        for (var i = currentAstPath.length - 2; i >= 0; i--) {
            if (getEntryName(currentAstPath[i]) === name) {
                return currentAstPath[i];
            }
        }
    }
    return null;
};

exports.getCurrentSrcLineNum = function(currentAstPath) {
    for (var i = currentAstPath.length - 1; i >= 0; i--) {
        if (typeof currentAstPath[i][0] === "object") {
            return currentAstPath[i][0].start.line;
        }
    }
    return 0;
};