var dns = require("dns"), tslErrorsDomain = require("domain").create(), http = require("http"), https = require("https"), os = require("os"), url = require("url"), util = require("util"), webauth = require("webauth"), Events = require("./events"), ServiceChannel = require("./service_channel"), ERR = require("./../server/server_errs"), exec = require("child_process").exec, injector = require("./../server/injector"), sharedConst = require("./../shared/const"), urlUtil = require("./../shared/url_util");

process.env.NODE_TLS_REJECT_UNAUTHORIZED = "0";

var httpAgent = new http.Agent(), httpsAgent = new https.Agent();

httpAgent.maxSockets = httpsAgent.maxSockets = 128;

httpsAgent.options.secureProtocol = "SSLv3_method";

var curUserDomain = "", curUserWorkstation = "", isWin = !!os.platform().match(/^win/);

if (isWin) {
    exec("echo %userdomain%", function(error, domain) {
        if (!error) {
            curUserDomain = domain.replace(/\s/g, "");
        }
    });
    exec("hostname", function(error, hostname) {
        if (!error) {
            curUserWorkstation = hostname.replace(/\s/g, "");
        }
    });
}

function getContentLength(bodyChunks) {
    var length = 0;
    bodyChunks.forEach(function(chunk) {
        length += chunk.length;
    });
    return length;
}

function testForPageMime(header) {
    return header && /(text\/html)|(application\/xhtml\+xml)|(application\/xml)|(application\/x-ms-application)/gi.test(header);
}

function getCharset(contentTypeHeader) {
    if (contentTypeHeader) {
        var charsetMatch = contentTypeHeader.match(/(?:^|;)\s*charset=(.+)(?:;|$)/i);
        if (charsetMatch) {
            return charsetMatch[1];
        }
    }
    return "iso-8859-1";
}

tslErrorsDomain.on("error", function(err) {
    if (err.code !== "ECONNRESET") {
        tslErrorsDomain.removeAllListeners("error");
        throw new Error(err);
    }
});

var Proxy = module.exports = function(port, hostname, cookieShelf) {
    this.port = port;
    this.hostname = hostname;
    this.events = new Events();
    this.cookieShelf = cookieShelf;
    this.serviceChannel = new ServiceChannel(this.port, this.hostname, this.cookieShelf);
    var proxy = this;
    this.server = http.createServer(function(req, res) {
        var parsedProxyUrl = urlUtil.parseProxyUrl(req.url);
        var isXhr = !!req.headers[sharedConst.XHR_REQUEST_MARKER_HEADER], isIFrame = parsedProxyUrl && parsedProxyUrl.isIFrame, ctx = {
            jobInfo: null,
            req: req,
            reqBodyChunks: [],
            res: res,
            resHeaders: {},
            originResourceInfo: null,
            originResContentInfo: null,
            originReqOptions: {},
            originReq: null,
            originRes: null,
            originResBodyChunks: [],
            isXhr: isXhr,
            isIFrame: isIFrame,
            isPageReqCandidate: !isXhr && testForPageMime(req.headers["accept"]),
            originReqErr: false
        };
        proxy._onProxyRequest(ctx);
    });
    this.server.on("error", function(err) {
        if (err.code === "EADDRINUSE") {
            proxy.events.broadcast.emit("fatalError", {
                code: ERR.PROXY_PORT_IS_ALREADY_IN_USE,
                port: proxy.port
            });
        }
    });
};

Proxy.prototype.start = function() {
    this.server.listen(this.port);
};

Proxy.prototype.close = function() {
    this.server.close();
};

Proxy.prototype.getProxyUrl = function(originUrl, jobUid, jobOwnerToken, isIFrame) {
    return urlUtil.getProxyUrl(originUrl, this.hostname, this.port, jobUid, jobOwnerToken, isIFrame);
};

Proxy.prototype.getResourceUrlReplacer = function(ctx) {
    var proxy = this;
    return function(resourceUrl, isIFrame) {
        var baseUrl = urlUtil.getBaseUrl();
        baseUrl = baseUrl ? url.resolve(ctx.originResourceInfo.url, baseUrl) : "";
        var resolvedUrl = url.resolve(baseUrl || ctx.originResourceInfo.url, resourceUrl);
        try {
            return proxy.getProxyUrl(resolvedUrl, ctx.jobInfo.uid, ctx.jobInfo.ownerToken, isIFrame);
        } catch (err) {
            return resourceUrl;
        }
    };
};

Proxy.prototype._error = function(ctx, err) {
    var eventsInCtx = this.events.in(ctx);
    if (eventsInCtx.canHandle("error")) {
        eventsInCtx.emit("error", err);
    } else {
        ctx.res.statusCode = 500;
        ctx.res.end();
    }
};

Proxy.prototype._extractReqInfo = function(ctx) {
    if (ctx.req.url === "/favicon.ico") {
        ctx.res.statusCode = 404;
        ctx.res.end();
        return;
    }
    var parsingResult = urlUtil.parseProxyUrl(ctx.req.url), referer = ctx.req.headers["referer"], refererParsingResult = null;
    try {
        refererParsingResult = referer && urlUtil.parseProxyUrl(referer);
    } catch (err) {
        refererParsingResult = null;
    }
    if (!parsingResult) {
        if (!refererParsingResult) {
            return null;
        }
        var parsedReqUrl = url.parse(ctx.req.url, true), hostParts = parsedReqUrl.host && parsedReqUrl.host.split(":");
        parsingResult = {
            originResourceInfo: {
                protocol: parsedReqUrl.protocol || refererParsingResult.originResourceInfo.protocol,
                host: parsedReqUrl.host || refererParsingResult.originResourceInfo.host,
                hostname: hostParts ? hostParts[0] : refererParsingResult.originResourceInfo.hostname,
                port: hostParts ? hostParts[1] : refererParsingResult.originResourceInfo.port,
                pathname: parsedReqUrl.pathname,
                query: parsedReqUrl.query,
                hash: parsedReqUrl.hash
            },
            jobInfo: refererParsingResult.jobInfo
        };
    }
    parsingResult.originResourceInfo.url = url.format(parsingResult.originResourceInfo);
    parsingResult.originResourceInfo.domain = url.format({
        protocol: parsingResult.originResourceInfo.protocol,
        host: parsingResult.originResourceInfo.host,
        hostname: parsingResult.originResourceInfo.hostname,
        port: parsingResult.originResourceInfo.port
    });
    if (refererParsingResult) {
        var refererResourceInfo = refererParsingResult.originResourceInfo;
        parsingResult.originResourceInfo.referer = url.format(refererResourceInfo);
        parsingResult.originResourceInfo.reqOrigin = url.format({
            protocol: refererResourceInfo.protocol,
            host: refererResourceInfo.host,
            hostname: refererResourceInfo.hostname,
            port: refererResourceInfo.port
        });
    }
    return parsingResult;
};

Proxy.prototype._onProxyRequest = function(ctx) {
    ctx.req.on("data", function(chunk) {
        ctx.reqBodyChunks.push(chunk);
    });
    var proxy = this;
    ctx.req.on("end", function() {
        if (proxy.serviceChannel.shouldProcess(ctx)) {
            proxy.serviceChannel.process(ctx);
        } else {
            proxy._lookupOrigin(ctx);
        }
    });
};

Proxy.prototype._lookupOrigin = function(ctx) {
    var requestInfo = this._extractReqInfo(ctx);
    if (!requestInfo) {
        ctx.res.statusCode = 404;
        ctx.res.end();
        return;
    }
    ctx.originResourceInfo = requestInfo.originResourceInfo;
    ctx.jobInfo = requestInfo.jobInfo;
    if (ctx.originResourceInfo.hostname === "localhost" || ctx.originResourceInfo.hostname === "127.0.0.1") {
        ctx.originReqOptions.hostname = "127.0.0.1";
        this._prepareOriginRequest(ctx);
    } else {
        var proxy = this;
        dns.lookup(ctx.originResourceInfo.hostname, function(err, addr) {
            if (err || !addr) {
                proxy._error(ctx, {
                    code: ERR.PROXY_CANT_RESOLVE_ORIGIN_URL,
                    originUrl: ctx.originResourceInfo.url
                });
                return;
            }
            ctx.originReqOptions.hostname = addr;
            proxy._prepareOriginRequest(ctx);
        });
    }
};

Proxy.prototype._prepareOriginRequest = function(ctx) {
    ctx.originReqOptions.port = ctx.originResourceInfo.port;
    ctx.originReqOptions.method = ctx.req.method;
    ctx.originReqOptions.path = urlUtil.formatUrl({
        pathname: ctx.originResourceInfo.pathname,
        query: ctx.originResourceInfo.query
    }, ctx.req.url);
    ctx.originReqOptions.hash = ctx.originResourceInfo.hash;
    ctx.originReqOptions.headers = {};
    ctx.originReqOptions.agent = ctx.originResourceInfo.protocol === "https:" ? httpsAgent : httpAgent;
    var proxy = this;
    Object.keys(ctx.req.headers).forEach(function(field) {
        switch (field) {
          case "host":
            ctx.originReqOptions.headers[field] = ctx.originResourceInfo.host;
            break;

          case "referer":
            ctx.originReqOptions.headers[field] = ctx.originResourceInfo.referer || ctx.req.headers["referer"];
            break;

          case "origin":
            ctx.originReqOptions.headers[field] = ctx.originResourceInfo.reqOrigin || ctx.req.headers["origin"];
            break;

          case "cookie":
            break;

          case sharedConst.XHR_REQUEST_MARKER_HEADER:
            break;

          default:
            ctx.originReqOptions.headers[field] = ctx.req.headers[field];
        }
    });
    if (ctx.isXhr && !ctx.req.headers["origin"] && ctx.originResourceInfo.domain !== ctx.originResourceInfo.reqOrigin) {
        ctx.originReqOptions.headers["origin"] = ctx.originResourceInfo.reqOrigin;
    }
    this.cookieShelf.getCookieHeader(ctx.jobInfo, ctx.originResourceInfo.url, function(cookieHeader) {
        ctx.originReqOptions.headers["cookie"] = cookieHeader;
        proxy._sendOriginRequest(ctx);
    });
};

Proxy.prototype._sendOriginRequest = function(ctx) {
    var proxy = this, isHttps = ctx.originResourceInfo.protocol === "https:", protocolInterface = isHttps ? https : http;
    if (isHttps) {
        ctx.originReqOptions.rejectUnauthorized = false;
    }
    tslErrorsDomain.run(function() {
        ctx.originReq = protocolInterface.request(ctx.originReqOptions, function(originRes) {
            ctx.originRes = originRes;
            if (originRes.statusCode === 401) {
                proxy.events.in(ctx).emit("authCredentialsRequested", function(credentials) {
                    if (credentials) {
                        ctx.originReq.abort();
                        webauth.auth(ctx.originReqOptions, credentials, function(targetRes) {
                            ctx.originRes = targetRes;
                            proxy._onOriginResponse(ctx);
                        }, isHttps, originRes);
                    } else {
                        proxy._onOriginResponse(ctx);
                    }
                });
            } else {
                proxy._onOriginResponse(ctx);
            }
        });
    });
    ctx.originReq.on("error", function() {
        ctx.originReqErr = true;
    });
    ctx.reqBodyChunks.forEach(function(chunk) {
        ctx.originReq.write(chunk);
    });
    ctx.originReq.end();
};

Proxy.prototype._isSameOriginPolicyCompliantRequest = function(ctx) {
    var reqOrigin = ctx.originResourceInfo.reqOrigin;
    if (ctx.originResourceInfo.domain !== reqOrigin) {
        if (ctx.req.headers[sharedConst.XHR_REQUEST_MARKER_HEADER] === sharedConst.XHR_CORS_SUPPORTED_FLAG) {
            if (ctx.req.method === "OPTIONS") {
                return true;
            }
            var allowOriginHeader = ctx.originRes.headers["access-control-allow-origin"], credentials = !!ctx.originRes.headers["access-control-allow-credentials"];
            return util.isArray(allowOriginHeader) ? allowOriginHeader.indexOf("*") > -1 && !credentials || allowOriginHeader.indexOf(reqOrigin) > -1 : allowOriginHeader === "*" && !credentials || allowOriginHeader === reqOrigin;
        }
        return false;
    }
    return true;
};

Proxy.prototype._onOriginResponse = function(ctx) {
    ctx.originResContentInfo = {
        encoding: ctx.originRes.headers["content-encoding"],
        charset: getCharset(ctx.originRes.headers["content-type"]),
        isPage: ctx.isPageReqCandidate && testForPageMime(ctx.originRes.headers["content-type"])
    };
    ctx.originRes.on("data", function(chunk) {
        ctx.originResBodyChunks.push(chunk);
    });
    if (ctx.isXhr && !this._isSameOriginPolicyCompliantRequest(ctx)) {
        this._error(ctx, {
            code: ERR.PROXY_XHR_REQUEST_SAME_ORIGIN_POLICY_VIOLATION,
            reqOrigin: ctx.originResourceInfo.reqOrigin,
            xhrReqUrl: ctx.originResourceInfo.url
        });
        ctx.res.statusCode = 0;
        ctx.res.end();
        return;
    }
    var proxy = this;
    ctx.originRes.on("end", function() {
        var cookies = ctx.originRes.headers["set-cookie"] || "";
        proxy.cookieShelf.setCookieByServer(ctx.jobInfo, ctx.originResourceInfo.url, cookies, function() {
            proxy._prepareProxyResponse(ctx);
        });
    });
};

Proxy.prototype._processResourceContent = function(ctx, callback) {
    if (ctx.originRes.statusCode === 200) {
        var urlReplacer = this.getResourceUrlReplacer(ctx), contentType = ctx.originRes.headers["content-type"] || "", accept = ctx.req.headers["accept"] || "", isCssResource = function() {
            return contentType.indexOf("text/css") !== -1 || accept === "text/css";
        }, isScriptResource = function() {
            return contentType.indexOf("application/javascript") !== -1 || contentType.indexOf("text/javascript") !== -1 || contentType.indexOf("application/x-javascript") !== -1 || accept === "application/javascript" || accept === "text/javascript" || accept === "application/x-javascript";
        }, isManifest = function() {
            return contentType.indexOf("text/cache-manifest") !== -1;
        }, isJSON = function() {
            return contentType.indexOf("application/json") !== -1;
        };
        if (isCssResource()) {
            injector.injectInStylesheet(ctx.originResBodyChunks, ctx.originResContentInfo.encoding, ctx.originResContentInfo.charset, urlReplacer, function(err, processedStylesheet) {
                if (!err) {
                    ctx.originResBodyChunks = [ processedStylesheet ];
                }
                callback();
            });
        } else {
            if (isScriptResource()) {
                injector.injectInScript(ctx.originResBodyChunks, ctx.originResContentInfo.encoding, ctx.originResContentInfo.charset, function(err, processedScript) {
                    if (!err) {
                        ctx.originResBodyChunks = [ processedScript ];
                    }
                    callback();
                });
            } else {
                if (isManifest()) {
                    injector.injectInManifest(ctx.originResBodyChunks, ctx.originResContentInfo.encoding, ctx.originResContentInfo.charset, urlReplacer, function(err, processedManifest) {
                        if (!err) {
                            ctx.originResBodyChunks = [ processedManifest ];
                        }
                        callback();
                    });
                } else {
                    if (isJSON()) {
                        injector.injectInJSON(ctx.originResBodyChunks, ctx.originResContentInfo.encoding, ctx.originResContentInfo.charset, function(err, processedJSON) {
                            if (!err) {
                                ctx.originResBodyChunks = [ processedJSON ];
                            }
                            callback();
                        });
                    } else {
                        callback();
                    }
                }
            }
        }
    } else {
        callback();
    }
};

Proxy.prototype._hasOriginReqError = function(ctx) {
    if (!ctx.originReqErr) {
        return false;
    }
    var contentLength = ctx.originRes.headers["content-length"];
    return !ctx.originRes || Object.keys(ctx.originRes.headers).length === 0 || typeof contentLength !== "undefined" && contentLength !== getContentLength(ctx.originResBodyChunks);
};

Proxy.prototype._prepareProxyResponse = function(ctx) {
    if (this._hasOriginReqError(ctx)) {
        this._error(ctx, {
            code: ERR.PROXY_ORIGIN_SERVER_CONNECTION_TERMINATED,
            originUrl: ctx.originResourceInfo.url
        });
        return;
    }
    var proxy = this;
    Object.keys(ctx.originRes.headers).forEach(function(field) {
        switch (field) {
          case "set-cookie":
          case "content-security-policy":
          case "x-content-security-policy":
          case "x-webkit-csp":
          case "access-control-allow-origin":
            return;

          case "location":
            var location = ctx.originRes.headers[field];
            var parsedLocation = url.parse(location);
            if (!parsedLocation.host) {
                location = url.resolve(ctx.originResourceInfo.url, location);
            }
            ctx.resHeaders[field] = proxy.getProxyUrl(location, ctx.jobInfo.uid, ctx.jobInfo.ownerToken);
            return;

          default:
            ctx.resHeaders[field] = ctx.originRes.headers[field];
        }
    });
    this._processResourceContent(ctx, function() {
        var completeProxyResponse = function() {
            ctx.resHeaders["content-length"] = getContentLength(ctx.originResBodyChunks);
            proxy._sendProxyResponse(ctx);
        };
        var eventsInCtx = proxy.events.in(ctx);
        if (eventsInCtx.canHandle("beforeProxyResponse")) {
            eventsInCtx.emit("beforeProxyResponse", completeProxyResponse);
        } else {
            completeProxyResponse();
        }
    });
};

Proxy.prototype._sendProxyResponse = function(ctx) {
    ctx.res.writeHeader(ctx.originRes.statusCode, ctx.resHeaders);
    ctx.res.addTrailers(ctx.originRes.trailers);
    try {
        ctx.originResBodyChunks.forEach(function(chunk) {
            ctx.res.write(chunk);
        });
        ctx.res.end();
    } catch (err) {}
};