var Events = require("./events"), urlUtil = require("../shared/url_util"), ServiceMsgCmd = require("../shared/service_msg_cmd"), uuid = require("node-uuid");

var URL_PATH_PREFIX = "/tcs", TEST_CAFE_SCRIPT_PATHNAME = "/testcafe.js", SERVICE_MSGS_PATHNAME = "/tccservicemsg";

var ServiceChannel = module.exports = function(proxyPort, proxyHostname, cookieShelf) {
    this.events = new Events();
    this.cookieShelf = cookieShelf;
    this.port = proxyPort;
    this.hostname = proxyHostname;
    this.staticResourceDescriptors = {};
    this.urlPathPrefix = URL_PATH_PREFIX + uuid.v4();
    this.serviceMsgUrl = this._getServiceUrl(SERVICE_MSGS_PATHNAME);
    this.testCafeScriptUrl = this._getServiceUrl(TEST_CAFE_SCRIPT_PATHNAME);
};

ServiceChannel.prototype._getServiceUrl = function(resourcePathname) {
    return urlUtil.formatUrl({
        protocol: "http:",
        hostname: this.hostname,
        port: this.port,
        pathname: this.urlPathPrefix + resourcePathname
    });
};

ServiceChannel.prototype._processServiceMsg = function(ctx) {
    var msgJson = "";
    ctx.reqBodyChunks.forEach(function(chunk) {
        msgJson += chunk.toString();
    });
    try {
        var msg = JSON.parse(msgJson);
    } catch (err) {
        ctx.res.statusCode = 500;
        ctx.res.end();
        return;
    }
    var sendServiceResponse = function(data) {
        if (data !== null && typeof data !== "undefined") {
            ctx.res.end(JSON.stringify(data));
        } else {
            ctx.res.end();
        }
    };
    if (msg.cmd === ServiceMsgCmd.SET_COOKIE) {
        var serviceChannel = this, reqInfo = urlUtil.parseProxyUrl(msg.url), originUrl = reqInfo ? urlUtil.formatUrl(reqInfo.originResourceInfo) : msg.url, jobInfo = {
            uid: msg.jobUid,
            ownerToken: msg.jobOwnerToken
        };
        this.cookieShelf.setCookieByClient(jobInfo, originUrl, msg.cookie, function() {
            serviceChannel.cookieShelf.getClientCookieString(jobInfo, originUrl, sendServiceResponse);
        });
    } else {
        this.events.for(msg.jobOwnerToken).emit("serviceMsg", msg, sendServiceResponse);
    }
};

ServiceChannel.prototype.shouldProcess = function(ctx) {
    var parsedReqUrl = urlUtil.parseUrl(ctx.req.url);
    return parsedReqUrl.pathname.indexOf(this.urlPathPrefix) === 0;
};

ServiceChannel.prototype.process = function(ctx) {
    var parsedReqUrl = urlUtil.parseUrl(ctx.req.url), resourcePathname = parsedReqUrl.pathname.replace(this.urlPathPrefix, "");
    if (resourcePathname === SERVICE_MSGS_PATHNAME) {
        this._processServiceMsg(ctx);
    } else {
        if (resourcePathname === TEST_CAFE_SCRIPT_PATHNAME) {
            var referer = ctx.req.headers["referer"];
            if (referer) {
                var refererInfo = urlUtil.parseProxyUrl(referer);
                refererInfo.originResourceInfo.url = urlUtil.formatUrl(refererInfo.originResourceInfo);
                this.events.for(refererInfo.jobInfo.ownerToken).emit("testCafeScriptRequested", refererInfo, function(script) {
                    ctx.res.setHeader("content-type", "text/javascript");
                    ctx.res.setHeader("cache-control", "no-cache, no-store, must-revalidate");
                    ctx.res.setHeader("pragma", "no-cache");
                    ctx.res.end(script);
                });
            } else {
                ctx.res.statusCode = 404;
                ctx.res.end();
            }
        } else {
            var resourceDescr = this.staticResourceDescriptors[resourcePathname];
            if (resourceDescr) {
                ctx.res.setHeader("content-type", resourceDescr.contentType);
                ctx.res.setHeader("cache-control", "max-age=86400, public");
                ctx.res.end(resourceDescr.content);
            } else {
                ctx.res.statusCode = 404;
                ctx.res.end();
            }
        }
    }
};

ServiceChannel.prototype.registerStaticResource = function(resourcePathname, resourceDescr) {
    this.staticResourceDescriptors[resourcePathname] = resourceDescr;
    return this._getServiceUrl(resourcePathname);
};