var fs = require("fs"), path = require("path");

function readFile(filePath, binary) {
    var fileData = fs.readFileSync(path.join(__dirname, filePath));
    return binary ? fileData : fileData.toString();
}

var AssetsManager = module.exports = function(proxyServiceChannel) {
    this.proxyServiceChannel = proxyServiceChannel;
    this.updateClientUi();
    this.TEMPLATES = {};
    this.TEMPLATES.TEST_RUN = readFile("../_compiled_/client_runtime/test_run.jstmpl");
    this.TEMPLATES.TEST_RUN_ERR_HANDLER = readFile("../_compiled_/client_runtime/test_run_err_handler.jstmpl");
    this.TEMPLATES.RECORDING = readFile("../_compiled_/client_runtime/recording.jstmpl");
    this.TEMPLATES.RECORDING_ERR_PAGE = readFile("../_compiled_/client_runtime/recorder_server_err_page.jstmpl");
    this.TEMPLATES.RECORDING_AUTHENTICATION_PAGE = readFile("../_compiled_/client_runtime/authentication_page.jstmpl");
};

var CLIENT_UI_CSS_FILE_PATH = "../_compiled_/client_runtime/styles.css", CLIENT_UI_SPRITE_FILE_PATH = "../_compiled_/client_runtime/sprite.png", TEMPLATE_VAR_PATTERN = /"<@\s*(\S+)\s*@>"/g;

AssetsManager.prototype.updateClientUi = function() {
    var spriteData = readFile(CLIENT_UI_SPRITE_FILE_PATH, true), styleData = readFile(CLIENT_UI_CSS_FILE_PATH);
    var spriteUrl = this.proxyServiceChannel.registerStaticResource("/uisprite.png", {
        contentType: "image/png",
        deleteAfterRequest: false,
        content: spriteData
    });
    styleData = this.compileTemplate(styleData, {
        uiSpriteUrl: spriteUrl
    });
    this.uiStyleUrl = this.proxyServiceChannel.registerStaticResource("/uistyle.css", {
        contentType: "text/css",
        deleteAfterRequest: false,
        content: styleData
    });
};

AssetsManager.prototype.compileTemplate = function(template, vars) {
    return template.replace(TEMPLATE_VAR_PATTERN, function(subStr, arg) {
        return vars && typeof vars[arg] !== "undefined" ? vars[arg] : subStr;
    });
};