var path = require("path"), util = require("util"), http = require("http"), EventEmitter = require("events").EventEmitter, express = require("express"), io = require("socket.io"), ejs = require("ejs"), WorkerPool = require("./worker_pool").WorkerPool, errMsgBuilder = require("./err_msg_builder"), ERR = require("./server_errs");

var ControlPanel = exports.ControlPanel = function(port, testRunner, vfs, workerPool, reporter, api, recorder) {
    EventEmitter.call(this);
    this.app = express();
    this.appServer = http.createServer(this.app);
    this.port = port;
    this.vfs = vfs;
    this.testRunner = testRunner;
    this.workerPool = workerPool;
    this.reporter = reporter;
    this.api = api;
    this.recorder = recorder;
    this.socketIO = null;
    this.app.use(express.bodyParser()).use("/assets", express.static(path.join(__dirname, "../_compiled_/control_panel_ui/assets"))).set("view engine", "ejs").set("view options", {
        layout: false
    }).set("views", path.join(__dirname, "../_compiled_/control_panel_ui/views"));
    var controlPanel = this;
    this.appServer.on("error", function(err) {
        if (err.code === "EADDRINUSE") {
            controlPanel.emit("fatalError", {
                code: ERR.CONTROL_PANEL_PORT_IS_ALREADY_IN_USE,
                port: controlPanel.port
            });
        }
    });
};

util.inherits(ControlPanel, EventEmitter);

ControlPanel.CONNECTED_WORKER = "Connected worker";

ControlPanel.AUTO_CREATED = "Auto created worker";

ControlPanel.CONNECTED_WORKER_TYPE = "Connected";

ControlPanel.AUTO_CREATED_WORKER_TYPE = "Auto-created";

ControlPanel.prototype.start = function() {
    this.socketIO = io.listen(this.appServer, {
        log: false
    });
    this.appServer.listen(this.port);
    this._setupSockets();
    this._setupRoutes();
};

ControlPanel.prototype.close = function() {
    this.appServer.close();
};

ControlPanel.prototype._setupRoutes = function() {
    var controlPanel = this, GET = function(route, handler) {
        controlPanel.app.get(route, handler.bind(controlPanel));
    }, POST = function(route, handler) {
        controlPanel.app.post(route, handler.bind(controlPanel));
    }, ALL = function(route, handler) {
        controlPanel.app.all(route, handler.bind(controlPanel));
    };
    GET("/", this._indexPageHandler);
    GET("/*", this._cachingPreventionHandler);
    POST("/*", this._cachingPreventionHandler);
    GET("/workers/", this._workersPageHandler);
    GET("/workers", this._workersPageHandler);
    ALL("/tests/*/", this._testBrowserHandler);
    ALL("/tests/*", this._testBrowserHandler);
    ALL("/tests", this._testBrowserHandler);
    POST("/tests_run/", this._runTests);
    POST("/tests_run", this._runTests);
    GET("/worker/add/:name/", this._addWorker);
    GET("/worker/add/:name", this._addWorker);
    GET("/worker/idle/:name/", this._workerIdleHandler);
    GET("/worker/idle/:name", this._workerIdleHandler);
    GET("/worker/list/", this._listWorkers);
    GET("/worker/list", this._listWorkers);
    POST("/worker/heartbeat/", this._workerHeartbeatHandler);
    POST("/worker/heartbeat", this._workerHeartbeatHandler);
    GET("/tasks/:taskID/", this._getDetailTaskReport);
    GET("/tasks/:taskID", this._getDetailTaskReport);
    GET("/tasks/", this._listTaskReports);
    GET("/tasks", this._listTaskReports);
    GET("/tasks_rev/:uid/", this._getTaskReportRevision);
    GET("/tasks_rev/:uid", this._getTaskReportRevision);
    POST("/rename_dir/", this._renameDir);
    POST("/rename_dir", this._renameDir);
    POST("/delete_dir/", this._deleteDir);
    POST("/delete_dir", this._deleteDir);
    POST("/create_fixture/", this._createFixture);
    POST("/create_fixture", this._createFixture);
    POST("/create_fixture_in_dir/", this._createFixtureInDir);
    POST("/create_fixture_in_dir", this._createFixtureInDir);
    POST("/edit_fixture/", this._editFixture);
    POST("/edit_fixture", this._editFixture);
    POST("/delete_fixture/", this._deleteFixture);
    POST("/delete_fixture", this._deleteFixture);
    POST("/rename_test/", this._renameTest);
    POST("/rename_test", this._renameTest);
    POST("/delete_test/", this._deleteTest);
    POST("/delete_test", this._deleteTest);
    POST("/dirs_by_path/", this._getDirsByPath);
    POST("/dirs_by_path", this._getDirsByPath);
    POST("/set_tests_dir/", this._setTestsDir);
    POST("/set_tests_dir", this._setTestsDir);
    POST("/recording/start/", this._record);
    POST("/recording/start", this._record);
    ALL("*", this._page404Handler);
};

ControlPanel.prototype._cachingPreventionHandler = function(req, res, next) {
    res.set("cache-control", "no-cache, no-store, must-revalidate");
    next();
};

ControlPanel.prototype._indexPageHandler = function(req, res) {
    res.statusCode = 301;
    res.setHeader("Location", "/tests/");
    res.end();
};

ControlPanel.prototype._workersPageHandler = function(req, res) {
    var isUpdateRequest = req.param("workersUpdate"), currentWorker = req.param("worker") || "", workers = this.workerPool.getWorkersInfo(), connectedWorkerSuggestionName = this.workerPool.getWorkerNameFromUserAgent(req.headers["user-agent"]);
    Object.keys(workers).forEach(function(workerIndex) {
        workers[workerIndex].workerType = workers[workerIndex].workerType === WorkerPool.WORKER_TYPE.CONNECTED ? ControlPanel.CONNECTED_WORKER_TYPE : ControlPanel.AUTO_CREATED_WORKER_TYPE;
    });
    if (!isUpdateRequest) {
        res.render("workers", {
            workers: workers,
            connectedWorkerSuggestionName: connectedWorkerSuggestionName,
            currentWorker: currentWorker
        });
    } else {
        res.render("partial/workers_container", {
            workers: workers
        });
    }
};

ControlPanel.prototype._testBrowserHandler = function(req, res) {
    var slide = req.param("slide"), rebuild = req.param("rebuild"), completeRecordingUid = req.param("completeRecordingUid"), expandRowFilename = req.param("expand") || "", urlPath = req.params[0] || "", vfsPath = urlPath ? decodeURI(urlPath).split("/") : [], vfs = this.vfs, testsBrowser = null, isPage404 = false, errs = [], fixtureErrs = [];
    this.api.listDirectory(vfsPath, function(err, pathInfo) {
        if (err && vfsPath.length) {
            isPage404 = true;
        } else {
            if (!err) {
                if (pathInfo.buildErrs) {
                    for (var errIndex = 0, errsLength = pathInfo.buildErrs.length; errIndex < errsLength; errIndex++) {
                        var error = pathInfo.buildErrs[errIndex];
                        if (error.fixtureUid) {
                            if (!error.filename) {
                                fixtureErrs[error.fixtureUid] = [];
                                continue;
                            }
                            var errorMsg = errMsgBuilder.build(error), tmpBasePath = vfs.basePath.replace(/\//g, "\\");
                            if (fixtureErrs.hasOwnProperty(error.fixtureUid)) {
                                fixtureErrs[error.fixtureUid].push(errorMsg);
                            } else {
                                fixtureErrs[error.fixtureUid] = [ errorMsg ];
                            }
                            errs.push({
                                filename: error.filename.replace(tmpBasePath, ""),
                                errText: errorMsg
                            });
                        } else {
                            errs.push(errMsgBuilder.build(error));
                        }
                    }
                }
                testsBrowser = {
                    errs: fixtureErrs,
                    currentUrl: urlPath ? encodeURI([ "/tests/", urlPath, "/" ].join("")) : "/tests/",
                    pathInfo: pathInfo,
                    breadcrumbs: []
                };
                for (var i = 0, url = "/tests/"; i < vfsPath.length; i++) {
                    url = encodeURI([ url, vfsPath[i], "/" ].join(""));
                    testsBrowser.breadcrumbs.push({
                        name: vfsPath[i],
                        url: url
                    });
                }
            }
        }
        if (isPage404) {
            res.status(404);
            res.render("404");
            return;
        }
        var testsDir = vfs.getBasePath();
        if (slide) {
            res.render("partial/tests_browser", {
                testsBrowser: testsBrowser,
                testsDir: testsDir
            });
        } else {
            if (rebuild) {
                res.render("partial/tests_revision_container", {
                    testsBrowser: testsBrowser,
                    testsDir: testsDir,
                    errMsgs: errs.length ? errs : null
                });
            } else {
                res.render("tests", {
                    testsBrowser: testsBrowser,
                    testsDir: testsDir,
                    expandRowFilename: expandRowFilename,
                    errMsgs: errs.length ? errs : null,
                    completeRecordingUid: completeRecordingUid
                });
            }
        }
    });
};

ControlPanel.prototype._runTests = function(req, res) {
    var source = decodeURI(req.param("source")), workers = req.param("workers") || [], runTestErrs = [], returnUrl = req.param("location") || "", browsers = req.param("browsers") || [], emulateCursor = req.param("emulateCursor") !== "false", sourceType = req.param("sourceType"), api = this.api, options = {
        sourceType: sourceType,
        source: source,
        workers: workers,
        browsers: browsers,
        emulateCursor: emulateCursor
    }, onTaskStarted = null;
    if (!workers.length && !browsers.length) {
        var workerName = this.workerPool.createWorkerFromControlPanelWindow(returnUrl, req.headers["user-agent"]);
        options.workers.push(workerName);
        onTaskStarted = function(errs) {
            res.send({
                workerName: workerName,
                errs: errs ? errMsgBuilder.build(runTestErrs.concat(errs)) : []
            });
        };
    } else {
        onTaskStarted = function(errs, taskUid, workers) {
            var workerList = workers || [], status = "success";
            if (errs) {
                runTestErrs = runTestErrs.concat(errMsgBuilder.build(errs));
            }
            if (runTestErrs.length || !workerList.length) {
                status = workerList.length ? "with-errors" : "fail";
            }
            res.status(200);
            if (status !== "fail") {
                res.send({
                    status: status,
                    errs: runTestErrs,
                    taskUid: taskUid
                });
            } else {
                res.send({
                    status: status,
                    errs: runTestErrs
                });
            }
        };
    }
    api.runTests(options, onTaskStarted);
};

ControlPanel.prototype._addWorker = function(req, res) {
    var noRedirect = req.param("noRedirect"), autoCreated = req.param("auto_created") || "", name = req.params.name, location = null, returnUrl = req.param("returnUrl") || "", validName = "", err = "";
    try {
        if (!autoCreated) {
            validName = this.testRunner.workerPool.connectWorker(name, req.headers["user-agent"]);
        } else {
            validName = this.testRunner.workerPool.createWorkerInNewBrowserWindow(name, req.headers["user-agent"]);
        }
        location = [ "/worker/idle/", encodeURIComponent(validName) ].join("");
    } catch (e) {
        validName = null;
        err = errMsgBuilder.build(e);
    }
    if (err) {
        res.status(403);
        res.send(err);
        return;
    }
    if (noRedirect) {
        res.send(location);
    } else {
        res.status(302);
        res.set("location", location);
        res.end();
    }
};

ControlPanel.prototype._workerIdleHandler = function(req, res) {
    var restartTestRun = req.param("restartTestRun"), name = req.params.name;
    if (this.workerPool.get(name)) {
        if (restartTestRun) {
            this.testRunner.restartCurrentTestRun(name);
        }
        var workerType = this.workerPool.get(name).type;
        workerType = workerType === WorkerPool.WORKER_TYPE.CONNECTED ? ControlPanel.CONNECTED_WORKER : ControlPanel.AUTO_CREATED;
        res.render("worker_idle", {
            workerName: name,
            workerType: workerType
        });
    } else {
        res.status(404);
        res.render("404");
    }
};

ControlPanel.prototype._workerHeartbeatHandler = function(req, res) {
    var state = this.testRunner.getWorkerStatus(req.param("name"));
    if (state) {
        res.send(state);
    } else {
        res.status(204);
        res.end();
    }
};

ControlPanel.prototype._listWorkers = function(req, res) {
    var render = req.param("render"), browsers = this.api.config.browsers, workerNames = this.api.listConnectedWorkers();
    if (render) {
        res.render("partial/worker_list", {
            workerNames: workerNames,
            browsers: browsers
        });
    } else {
        res.send(workerNames);
    }
};

ControlPanel.prototype._getDetailTaskReport = function(req, res) {
    var taskID = req.params.taskID, partial = req.param("partial"), taskReport = this.reporter.getTaskReportByUid(taskID);
    if (taskReport) {
        if (partial) {
            res.render("partial/task_detail", {
                r: taskReport
            });
        } else {
            res.render("task", {
                r: taskReport
            });
        }
    } else {
        res.status(404);
        res.render("404");
    }
};

ControlPanel.prototype._listTaskReports = function(req, res) {
    var partial = req.param("partial");
    if (!partial) {
        res.render("task_list", {
            taskReports: this.reporter.getTaskReports()
        });
    } else {
        res.render("partial/tasks_report_container", {
            taskReports: this.reporter.getTaskReports()
        });
    }
};

ControlPanel.prototype._getTaskReportRevision = function(req, res) {
    var taskID = req.params.uid, isDetail = req.param("detail"), taskReport = this.reporter.getTaskReportByUid(taskID);
    if (taskReport) {
        if (isDetail) {
            res.render("partial/task_detail", {
                r: taskReport
            });
        } else {
            res.render("partial/task_list_item", {
                r: taskReport
            });
        }
    } else {
        res.status(404);
        res.render("404");
    }
};

ControlPanel.prototype._renameDir = function(req, res) {
    var urlPath = req.param("curPath") || "", vfsPath = urlPath ? decodeURI(urlPath).split("/") : [], newName = req.param("newName");
    this.vfs.renameDir(vfsPath, newName, function(err) {
        if (!err) {
            res.status(204);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

ControlPanel.prototype._deleteDir = function(req, res) {
    var urlPath = req.param("curPath") || "", vfsPath = urlPath ? decodeURI(urlPath).split("/") : [];
    this.vfs.deleteDir(vfsPath, function(err) {
        if (!err) {
            res.status(204);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

ControlPanel.prototype._createFixture = function(req, res) {
    var urlPath = req.param("curPath") || "", fixtureName = req.param("fixtureName"), fixturePage = req.param("fixturePage"), filename = req.param("filename"), fixtureUsername = req.param("fixtureUsername"), fixturePassword = req.param("fixturePassword"), vfsPath = urlPath ? decodeURI(urlPath).split("/") : [];
    this.vfs.createFixture(vfsPath, fixtureName, filename, fixturePage, fixtureUsername, fixturePassword, function(err) {
        if (!err) {
            res.status(204);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

ControlPanel.prototype._createFixtureInDir = function(req, res) {
    var urlPath = req.param("curPath") || "", fixtureName = req.param("fixtureName"), fixturePage = req.param("fixturePage"), fixtureUsername = req.param("fixtureUsername"), fixturePassword = req.param("fixturePassword"), filename = req.param("filename"), dirName = req.param("dirName"), vfsPath = urlPath ? decodeURI(urlPath).split("/") : [];
    this.vfs.createFixtureInDir(vfsPath, dirName, fixtureName, filename, fixturePage, fixtureUsername, fixturePassword, function(err) {
        if (!err) {
            res.status(204);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

ControlPanel.prototype._editFixture = function(req, res) {
    var urlPath = req.param("curPath") || "", newName = req.param("newName"), newPage = req.param("newPage"), oldFilename = req.param("oldFilename"), newFilename = req.param("newFilename"), newUsername = req.param("newUsername"), newPassword = req.param("newPassword"), vfsPath = urlPath ? decodeURI(urlPath).split("/") : [];
    this.vfs.editFixture(vfsPath, oldFilename, newFilename, newName, newPage, newUsername, newPassword, function(err) {
        if (!err) {
            res.status(204);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

ControlPanel.prototype._deleteFixture = function(req, res) {
    var urlPath = req.param("curPath") || "", fixtureFilename = req.param("fixtureFilename"), vfsPath = urlPath ? decodeURI(urlPath).split("/") : [];
    this.vfs.deleteFixture(vfsPath, fixtureFilename, function(err) {
        if (!err) {
            res.status(204);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

ControlPanel.prototype._deleteTest = function(req, res) {
    var urlPath = req.param("curPath") || "", filename = req.param("filename"), testName = req.param("testName"), vfsPath = urlPath ? decodeURI(urlPath).split("/") : [];
    this.vfs.deleteTest(vfsPath, filename, testName, function(err) {
        if (!err) {
            res.status(204);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

ControlPanel.prototype._renameTest = function(req, res) {
    var urlPath = req.param("curPath") || "", filename = req.param("filename"), oldName = req.param("oldName"), newName = req.param("newName"), vfsPath = urlPath ? decodeURI(urlPath).split("/") : [];
    this.vfs.renameTest(vfsPath, filename, oldName, newName, function(err) {
        if (!err) {
            res.status(204);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

ControlPanel.prototype._record = function(req, res) {
    var returnUrl = req.param("returnUrl"), fixtureUid = req.param("fixtureUid");
    this.recorder.startRecording(fixtureUid, returnUrl, function(err, recordingUrl) {
        if (err) {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        } else {
            res.send(recordingUrl);
        }
    });
};

ControlPanel.prototype._getDirsByPath = function(req, res) {
    var fsPath = req.param("path");
    if (fsPath) {
        this.vfs.getDirsByPath(decodeURIComponent(fsPath), function(err, dirs) {
            res.status(200);
            res.send(dirs);
        });
    } else {
        this.vfs.getLocalDrives(function(err, devicesList) {
            if (!err) {
                res.status(200);
                res.send(devicesList);
            } else {
                res.status(304);
                res.send(errMsgBuilder.build(err));
            }
        });
    }
};

ControlPanel.prototype._setTestsDir = function(req, res) {
    var fsPath = decodeURIComponent(req.param("testsDir"));
    this.vfs.setTestsDir(fsPath, function(err) {
        if (!err) {
            res.status(200);
            res.end();
        } else {
            res.status(403);
            res.send(errMsgBuilder.build(err));
        }
    });
};

ControlPanel.prototype._page404Handler = function(req, res) {
    res.status(404);
    res.render("404");
};

ControlPanel.prototype._setupSockets = function() {
    var socketIO = this.socketIO;
    this.testRunner.on("taskAdded", function(workerNames) {
        workerNames.forEach(function(workerName) {
            socketIO.sockets.in(workerName).emit("taskAdded");
        });
        socketIO.sockets.in("registerSeveralWorkers").emit("taskAdded");
    });
    this.reporter.on("taskUpdated", function(data) {
        socketIO.sockets.emit("taskUpdated", {
            id: data.id
        });
    });
    this.reporter.on("taskDetailErrUpdated", function(data) {
        socketIO.sockets.emit("taskDetailErrUpdated", {
            id: data.id
        });
    });
    this.reporter.on("taskListTestFailed", function(data) {
        socketIO.sockets.emit("taskListTestFailed", {
            id: data.id
        });
    });
    this.vfs.on("suiteChanged", function() {
        socketIO.sockets.emit("vfsRevisionChanged");
    });
    this.workerPool.on("workerAdded", function() {
        socketIO.sockets.emit("workerAdded");
    });
    this.workerPool.on("workerDisconnected", function() {
        socketIO.sockets.emit("workerDisconnected");
    });
    socketIO.sockets.on("connection", function(socket) {
        socket.on("registerWorker", function(workerName) {
            if (workerName && !socket.disconnected) {
                socket.join(workerName);
            }
        });
        socket.on("registerSeveralWorkers", function() {
            if (!socket.disconnected) {
                socket.join("workerList");
            }
        });
    });
};