var util = require("util"), url = require("url"), cookies = require("tough-cookie"), async = require("async");

function getJarUid(jobInfo) {
    return jobInfo.ownerToken + jobInfo.uid;
}

var CookieShelf = exports.CookieShelf = function() {
    this.shelf = {};
};

CookieShelf.prototype._getJar = function(jobInfo, createIfNotExist) {
    var jarUid = getJarUid(jobInfo), jar = this.shelf[jarUid];
    if (!jar && createIfNotExist) {
        jar = this.shelf[jarUid] = new cookies.CookieJar();
    }
    return jar;
};

CookieShelf.prototype._setCookie = function(jobInfo, resourceUrl, cookie, client, callback) {
    var jar = this._getJar(jobInfo, true);
    async.forEach(util.isArray(cookie) ? cookie : [ cookie ], function(cookieEntry, cookieSetCallback) {
        jar.setCookie(cookieEntry, resourceUrl, {
            http: !client,
            strict: true
        }, function() {
            cookieSetCallback();
        });
    }, callback);
};

CookieShelf.prototype._getCookieString = function(jobInfo, resourceUrl, client, callback) {
    var jar = this._getJar(jobInfo);
    if (jar) {
        jar.getCookieString(resourceUrl, {
            http: !client
        }, function(err, cookieStr) {
            callback(cookieStr);
        });
    } else {
        callback("");
    }
};

CookieShelf.prototype.setCookieByServer = function(jobInfo, resourceUrl, cookie, callback) {
    this._setCookie(jobInfo, resourceUrl, cookie, false, callback);
};

CookieShelf.prototype.setCookieByClient = function(jobInfo, resourceUrl, cookie, callback) {
    this._setCookie(jobInfo, resourceUrl, cookie, true, callback);
};

CookieShelf.prototype.getClientCookieString = function(jobInfo, resourceUrl, callback) {
    this._getCookieString(jobInfo, resourceUrl, true, callback);
};

CookieShelf.prototype.getCookieHeader = function(jobInfo, resourceUrl, callback) {
    this._getCookieString(jobInfo, resourceUrl, false, callback);
};

CookieShelf.prototype.removeCookies = function(jobInfo) {
    var jarUid = getJarUid(jobInfo);
    delete this.shelf[jarUid];
};