var url = require("url"), util = require("util"), FixtureCode = require("../fixture_code"), ERR = require("./server_errs"), CLIENT_ERR = require("./../shared/client_errs");

exports.UNKNOW_ERR_MSG = "UNKNOWN_ERROR";

var generateMsg = function(err) {
    var msg = exports.UNKNOW_ERR_MSG, m = function() {
        msg = util.format.apply(this, arguments);
    };
    switch (err.code) {
      case ERR.API_TESTS_TARGET_NOT_EXIST:
        m('The "%s" test target is empty or not found.', err.source);
        break;

      case ERR.API_UNKNOWN_FIXTURE_TYPE:
        m('The "%s" fixture type is unknown. The allowed fixture types are: "dir", "fixture", and "test".', err.type);
        break;

      case ERR.API_NO_TESTS_TO_RUN:
        m("There are no tests to run.");
        break;

      case ERR.API_TASK_WORKER_DOESNT_EXIST:
        m('Attempt to start the "%s" task on a non-existing worker "%s".', err.taskName, err.workerName);
        break;

      case ERR.API_NO_WORKERS_SPECIFIED_FOR_TASK:
        m("Specify either a browser or a worker to run the test.");
        break;

      case ERR.CONFIG_PROPERTY_CAN_NOT_BE_EMPTY:
        m('The "%s" property value cannot be empty. Please correct the "config.json" configuration file.', err.property);
        break;

      case ERR.CONFIG_PROPERTY_IS_NOT_STRING:
        m('The "%s" property value must be of a string type. Please correct the "config.json" configuration file.', err.property);
        break;

      case ERR.CONFIG_PROPERTY_IS_REQUIRED:
        m('The "%s" property is required. Please define the property in the "config.json" configuration file.', err.property);
        break;

      case ERR.CONFIG_UNSUPPORTED_PROPERTY:
        m('The "%s" property is not supported. Please remove the property from the "config.json" configuration file or check its spelling.', err.property);
        break;

      case ERR.CONFIG_PORT_VALUE_IS_NOT_NUMBER:
        m('The "%s" port property value must be of an integer type. Please correct the "config.json" configuration file.', err.property);
        break;

      case ERR.CONFIG_PORT_VALUE_IS_OUT_OF_RANGE:
        m('The "%s" port value must be between 0 and 65535. Please correct the "config.json" configuration file.', err.property);
        break;

      case ERR.CONFIG_INVALID_PROPERTY_FORMAT:
        m('The "%s" property value has incorrect format. Please check the property format in the "config.json" configuration file.', err.property);
        break;

      case ERR.CONFIG_READ_FILE_FAILED:
        m('Failed to read TestCafe configuration from the "config.json" file.');
        break;

      case ERR.CONFIG_PARSING_FAILED:
        m('Failed to parse a TestCafe configuration file ("config.json"). %s', err.parserErr);
        break;

      case ERR.CONTROL_PANEL_PORT_IS_ALREADY_IN_USE:
        m("Failed to start TestCafe Control Panel on the %s port. It seems that another application is now " + 'using this port. Use "config.json" file to modify TestCafe Control Panel port.', err.port);
        break;

      case FixtureCode.ErrCodes.READ_FILE_FAILED:
        m('Failed to read test file "%s."', err.filename);
        break;

      case FixtureCode.ErrCodes.JAVASCRIPT_PARSING_FAILED:
        m("(line %s): %s", err.parserErr && err.parserErr.line, err.parserErr && err.parserErr.message);
        break;

      case FixtureCode.ErrCodes.AUTH_DIRECTIVE_REDEFINITION:
        m("(line %s): @auth directive redefinition. @auth directive should be used once per test fixture file.", err.line);
        break;

      case FixtureCode.ErrCodes.FIXTURE_DIRECTIVE_REDEFINITION:
        m("(line %s): @fixture directive redefinition. @fixture directive should be used once per test fixture file.", err.line);
        break;

      case FixtureCode.ErrCodes.PAGE_DIRECTIVE_REDEFINITION:
        m("(line %s): @page directive redefinition. @page directive should be used once per test fixture file.", err.line);
        break;

      case FixtureCode.ErrCodes.INVALID_NETWORK_AUTHENTICATION_CREDENTIALS_FORMAT:
        m("(line %s): network authentication credentials have incorrect format. @auth directive should use the " + 'following format: "@auth username:password".', err.line);
        break;

      case FixtureCode.ErrCodes.REQUIRED_FILE_ALREADY_INCLUDED:
        m('(line %s): Required file "%s" is already included.', err.line, err.req);
        break;

      case FixtureCode.ErrCodes.MISPLACED_DIRECTIVE:
        m("(line %s): Misplaced directive. Directives (@fixture, @page, etc.) should be declared " + "on a global scope.", err.line);
        break;

      case FixtureCode.ErrCodes.MISPLACED_TEST_DECLARATION:
        m("(line %s): Misplaced test directive. Tests should be declared on a global scope.", err.line);
        break;

      case FixtureCode.ErrCodes.DUPLICATE_TEST_NAME:
        m('(line %s): Test with name "%s" is alredy defined.', err.line, err.testName);
        break;

      case FixtureCode.ErrCodes.TEST_NAME_CHANGED_TO_ALREADY_USED:
        m('Test with name "%s" is alredy defined.', err.testName);
        break;

      case FixtureCode.ErrCodes.EMPTY_TEST_NAME:
        m("(line %s): Test name is not set.", err.line);
        break;

      case FixtureCode.ErrCodes.INVALID_TEST_ASSIGNMENT:
        m("(line %s): Invalid test assignment. Test should be an object.", err.line);
        break;

      case FixtureCode.ErrCodes.TEST_STEP_IS_NOT_A_FUNCTION:
        m("(line %s): Test step is not a function.", err.line);
        break;

      case FixtureCode.ErrCodes.ACTION_FUNC_IS_NOT_A_LAST_ENTRY:
        m("(line %s): Action function call is not the last statement. Test step should not contain " + "any statements after a action function call.", err.line);
        break;

      case FixtureCode.ErrCodes.TEST_STEP_IS_EMPTY:
        m("(line %s): Test step doesn't contain any code.", err.line);
        break;

      case FixtureCode.ErrCodes.FIXTURE_DIRECTIVE_IS_UNDEFINED:
        m("%s: Fixture name is undefined. Test file should contain @fixture directive.", err.filename);
        break;

      case FixtureCode.ErrCodes.PAGE_DIRECTIVE_IS_UNDEFINED:
        m("%s: Page URL is undefined. Test file should contain @page directive.", err.filename);
        break;

      case FixtureCode.ErrCodes.ASYNC_FUNC_CALL:
        m("(line %s): Async function call. Test steps should not contain async functions calls, " + "use act.wait() instead.", err.line);
        break;

      case FixtureCode.ErrCodes.ACTION_FUNC_CALL_IN_SHARED_CODE:
        m("An action function cannot be called from the shared code. It can be used within a " + "test step only.", err.filename, err.line);
        break;

      case FixtureCode.ErrCodes.FAILED_LOAD_REQUIRE:
        m('%s: Failed load required file "%s".', err.filename, err.req);
        break;

      case FixtureCode.ErrCodes.MODULE_NOT_FOUND:
        m('(line %s): Required module "%s" does not exists.', err.line, err.moduleName);
        break;

      case FixtureCode.ErrCodes.TEST_IS_EMPTY:
        m("(line %s): Test doesn't contain any code.", err.line);
        break;

      case FixtureCode.ErrCodes.WRITE_FILE_FAILED:
        m('Failed to write test fixture code to the "%s" file.', err.filename);
        break;

      case FixtureCode.ErrCodes.INVALID_FILE_FORMAT:
        m('Failed to perform the operation because the "%s" test fixture file is not valid.', err.filename);
        break;

      case FixtureCode.ErrCodes.ELEMENT_SELECTOR_PARSING_FAILED:
        m("Failed to generate test code due to a syntax error in an element selector: %s", err.parserErr && err.parserErr.message);
        break;

      case FixtureCode.ErrCodes.TEST_IS_NOT_FOUND:
        m("Test %s is not found in this fixture", err.testName);
        break;

      case ERR.INJECTOR_RESOURCE_DECODING_FAILED:
        m('Failed to decode test page markup using %s decoder suggested by server response "Encoding" ' + "HTTP-header. This problem may occur if a test page server is misconfigured.", err.encoding);
        break;

      case ERR.INJECTOR_RESOURCE_CHARSET_DECODING_FAILED:
        m('Failed to decode test page markup charset using %s decoder that was suggested by server response "Content-type" ' + "HTTP-header. This problem may occur if a test page server is misconfigured.", err.charset);
        break;

      case ERR.INJECTOR_DOCUMENT_PARSING_FAILED:
        m("Failed to parse test page HTML-markup. %s", err.msg && err.msg.toString());
        break;

      case ERR.INJECTOR_RESOURCE_ENCODING_FAILED:
        m("TestCafe failed to encode proxy response body using %s encoder. The reason may be a bug in TestCafe " + "server implementation. Please contact the TestCafe vendor for details.", err.encoding);
        break;

      case ERR.PROXY_PORT_IS_ALREADY_IN_USE:
        m("Failed to start TestCafe service on port %s. TestCafe has already been started or another process " + 'is using this port. To resolve port conflicts, edit "config.json" file.', err.port);
        break;

      case ERR.PROXY_ORIGIN_SERVER_CONNECTION_TERMINATED:
        m('Failed to perform a request for the "%s" resource because origin server connection was unexpectedly ' + "terminated.", err.originUrl);
        break;

      case ERR.PROXY_CANT_RESOLVE_ORIGIN_URL:
        m('Failed to find DNS-record for the "%s" resource.', err.originUrl);
        break;

      case ERR.PROXY_XHR_REQUEST_SAME_ORIGIN_POLICY_VIOLATION:
        m('XMLHttpRequest from page "%s" to URL "%s" violates Same Origin Policy. XMLHttpRequest URL domain ' + "should match the current page domain", err.reqOrigin, err.xhrReqUrl);
        break;

      case ERR.RECORDER_TARGET_FIXTURE_DOESNT_EXISTS:
        m("The text fixture specified for test recording is not found.");
        break;

      case ERR.RECORDER_RECORDING_IS_EMPTY:
        m("Test doesn't contain any steps.");
        break;

      case ERR.RECORDER_TEST_NAME_IS_EMPTY:
        m("Test name is empty.");
        break;

      case ERR.RECORDER_RECORDING_DOESNT_EXISTS:
        m("Recording does not exist.");
        break;

      case ERR.SUITE_CFG_PROPERTY_IS_NOT_STRING:
        m('The "%s" property value must be of a string type. Please correct the "%s" file.', err.property, err.cfgPath);
        break;

      case ERR.SUITE_CFG_INVALID_PROPERTY_FORMAT:
        m('The "%s" property value has an incorrect format. Please check the property format in the "%s" file.', err.property, err.cfgPath);
        break;

      case ERR.SUITE_CFG_UNSUPPORTED_PROPERTY:
        m('The "%s" property is not supported. Please remove the property from the "%s" file or check its spelling.', err.property, err.cfgPath);
        break;

      case ERR.SUITE_READ_DIR_FAILED:
        m('Failed to read directory "%s"', err.dirPath);
        break;

      case ERR.SUITE_GET_FILE_STATS_FAILED:
        m('Failed to obtain file system statistics for path "%s"', err.filePath);
        break;

      case ERR.SUITE_READ_CFG_FILE_FAILED:
        m('Failed to read the "test_config.json" file at path "%s"', err.dirPath);
        break;

      case ERR.SUITE_CFG_FILE_PARSING_FAILED:
        m('Failed to parse the "test_config.json" file at path "%s". %s', err.dirPath, err.msg);
        break;

      case ERR.SUITE_CFG_MODULE_IS_NOT_ARRAY:
        m('Failed to parse %s file. Module "%s" is not an array of files.', err.cfgPath, err.property);
        break;

      case ERR.TEST_RUNNER_RESOURCE_ERROR_HTTP_STATUS_CODE:
        m('Failed to load page resource "%s". HTTP status code - %s.', err.url, err.statusCode);
        break;

      case ERR.TEST_RUNNER_MAX_TEST_RUN_RESTARTS:
        m("Failed to run the test since the test page loading hangs repetitively.");
        break;

      case ERR.VIRTUAL_FILE_SYSTEM_BASE_PATH_DOESNT_EXISTS:
        m('Tests directory path "%s", which is set in file "config.json", does not exist.', err.basePath);
        break;

      case ERR.VIRTUAL_FILE_SYSTEM_BASE_PATH_IS_NOT_DIR:
        m('Tests directory path "%s", which is set in file "config.json", does not exist.', err.basePath);
        break;

      case ERR.VIRTUAL_FILE_SYSTEM_CREATE_DIR_FAILED:
        m('Failed to create the "%s" directory.', err.dirPath);
        break;

      case ERR.VIRTUAL_FILE_SYSTEM_INVALID_DIR_PATH:
        m('The "%s" directory name is not valid.', err.dirPath);
        break;

      case ERR.VIRTUAL_FILE_SYSTEM_RENAME_DIR_FAILED:
        m('Failed to rename the "%s" directory.', err.dirPath);
        break;

      case ERR.VIRTUAL_FILE_SYSTEM_DELETE_DIR_FAILED:
        m('Failed to delete the "%s" directory.', err.dirPath);
        break;

      case ERR.VIRTUAL_FILE_SYSTEM_DIR_ALREADY_EXISTS:
        m('The "%s" directory already exists.', err.dirPath);
        break;

      case ERR.VIRTUAL_FILE_SYSTEM_DIR_DOESNT_EXIST:
        m('The "%s" directory does not exist.', err.dirPath);
        break;

      case ERR.VIRTUAL_FILE_SYSTEM_INVALID_FIXTURE_PATH:
        m('The "%s" file name is not valid.', err.filePath);
        break;

      case ERR.VIRTUAL_FILE_SYSTEM_FIXTURE_FILE_ALREADY_EXISTS:
        m("File already exists.");
        break;

      case ERR.VIRTUAL_FILE_SYSTEM_RENAME_FIXTURE_FILE_FAILED:
        m('Failed to rename the "%s" fixture.', err.filePath);
        break;

      case ERR.VIRTUAL_FILE_SYSTEM_FIXTURE_FILE_DOESNT_EXIST:
        m('The "%s" file does not exist.', err.filePath);
        break;

      case ERR.VIRTUAL_FILE_SYSTEM_DELETE_FIXTURE_FILE_FAILED:
        m('Failed to delete the "%s" file.', err.filePath);
        break;

      case ERR.VIRTUAL_FILE_SYSTEM_DIR_DOESNT_READ:
        m('The "%s" directory can\'t be read. It might have been deleted.', err.fsPath);
        break;

      case ERR.VIRTUAL_FILE_SYSTEM_NODEJS_CANT_GET_DRIVE_LIST:
        m("It is impossible to get the list of local directories.");
        break;

      case ERR.WORKER_POOL_WORKER_NAME_IS_EMPTY:
        m("Worker name is empty. The name should contain at least one character which cannot be a blank space.");
        break;

      case ERR.WORKER_POOL_WORKER_NAME_IS_ALREADY_USED:
        m('Worker with name "%s" is already registered. Please choose another name.', err.name);
        break;

      case ERR.WORKER_POOL_WORKER_DISCONNECTED:
        m('The "%s" worker did not respond to the TestCafe and was deleted. This problem may appear when a ' + "worker browser hangs or is closed, or due to network issues.", err.name);
        break;

      case ERR.WORKER_POOL_FAILED_TO_START_BROWSER:
        m('Failed to start browser "%s" located at path "%s". This may occur if the browser is not ' + 'configured well in the "config.json" file.', err.browserName, err.path);
        break;

      case ERR.WORKER_POOL_FAILED_TO_CREATE_WORKER_IN_BROWSER_WINDOW:
        m('Failed to start a worker on browser "%s". This may occur if the browser was closed before the ' + "worker initialization or due to browser hangup.", err.browserName);
        break;

      case CLIENT_ERR.URL_UTIL_PROTOCOL_IS_NOT_SUPPORTED:
        m('Failed to process the "%s" resource. TestCafe supports only HTTP and HTTPS protocols.', err.originUrl);
        break;

      case CLIENT_ERR.UNCAUGHT_JS_ERROR:
        m('Uncaught JavaScript error - "%s"', err.scriptErr);
        break;

      case CLIENT_ERR.UNCAUGHT_JS_ERROR_IN_TEST_CODE_STEP:
        m('Error on step "%s": Uncaught JavaScript error in test code - "%s"', err.stepName, err.scriptErr);
        break;

      case CLIENT_ERR.TEST_INACTIVITY:
        m("Abnormal test inactivity encountered. It may be caused by a critical error on a test page.");
        break;

      case CLIENT_ERR.STORE_DOM_NODE_OR_JQUERY_OBJECT:
        m('Error on step "%s": It is not allowed to share the DOM-element, jQuery object or a function between test steps via "this" object.', err.stepName);
        break;

      case CLIENT_ERR.XHR_REQUEST_TIMEOUT:
        m("Abnormally long XMLHttpRequest encountered.");
        break;

      case CLIENT_ERR.IFRAME_LOADING_TIMEOUT:
        m("Abnormally long iframe loading encountered.");
        break;

      case CLIENT_ERR.API_EMPTY_FIRST_ARGUMENT:
        m('Error on step "%s": "%s" action target does not contain DOM-elements.', err.stepName, err.action);
        break;

      case CLIENT_ERR.API_INVISIBLE_ACTION_ELEMENT:
        m('Error on step "%s": "%s" action target element ("%s") is not visible.', err.stepName, err.action, err.element);
        break;

      case CLIENT_ERR.API_INCORRECT_DRAGGING_SECOND_ARGUMENT:
        m('Error on step "%s": "%s" action drop target is incorrect.', err.stepName, err.action || "", err.element || "");
        break;

      case CLIENT_ERR.API_INCORRECT_PRESS_ACTION_ARGUMENT:
        m('Error on step "%s": press action parameter contains incorrect key code', err.stepName);
        break;

      case CLIENT_ERR.API_EMPTY_TYPE_ACTION_ARGUMENT:
        m('Error on step "%s": the type action\'s parameter text is empty.', err.stepName);
        break;

      case CLIENT_ERR.API_OK_ASSERTION_FAILED:
        if (typeof err.message !== "undefined" && err.message !== null) {
            m('"%s" assertion failed on step "%s". Expected value is a non-null value, true-boolean ' + 'value or non-empty string, but the actual is "%s".', err.message, err.stepName, err.actual);
        } else {
            m('Assertion failed on step "%s". Expected value is a non-null value, true-boolean value ' + 'or non-empty string, but the actual is "%s".', err.stepName, err.actual);
        }
        break;

      case CLIENT_ERR.API_EQUAL_ASSERTION_FAILED:
        var messagePrefix = typeof err.message !== "undefined" && err.message !== null ? err.message + "assertion" : "Assertion";
        if (err.isArrays) {
            m('"%s" failed on step "%s". Arrays differ in the "%s" index. The expected value is "%s", while the actual value is "%s".', messagePrefix, err.stepName, err.key, err.expected, err.actual);
        } else {
            if (err.isObjects) {
                m('"%s" failed on step "%s". Objects differ in the "%s" field. The expected value is "%s", while the actual value is "%s".', messagePrefix, err.stepName, err.key, err.expected, err.actual);
            } else {
                m('"%s" failed on step "%s". The expected value is "%s", while the actual value is "%s".', messagePrefix, err.stepName, err.expected, err.actual);
            }
        }
        break;

      case CLIENT_ERR.API_NOT_OK_ASSERTION_FAILED:
        if (typeof err.message !== "undefined" && err.message !== null) {
            m('"%s" assertion failed on step "%s". Expected value is a null value, false-boolean ' + 'value or empty string, but the actual is "%s".', err.message, err.stepName, err.actual);
        } else {
            m('Assertion failed on step "%s". Expected value is a null value, false-boolean value' + ' or empty string, but the actual is "%s".', err.stepName, err.actual);
        }
        break;

      case CLIENT_ERR.API_NOT_EQUAL_ASSERTION_FAILED:
        if (typeof err.message !== "undefined" && err.message !== null) {
            m('"%s" assertion failed on step "%s". Value "%s" was not expected.', err.message, err.stepName, err.actual);
        } else {
            m('Assertion failed on step "%s". Value "%s" was not expected.', err.stepName, err.actual);
        }
        break;

      case CLIENT_ERR.API_UNEXPECTED_DIALOG:
        m('Error on step "%s": unexpected system %s dialog "%s" appeared', err.stepName, err.dialog, err.message);
        break;

      case CLIENT_ERR.API_EXPECTED_DIALOG_DOESNT_APPEAR:
        m('Error on step "%s": the expected system %s dialog did not appear', err.stepName, err.dialog);
        break;

      case CLIENT_ERR.API_INCORRECT_SELECT_ACTION_ARGUMENTS:
        m('Error on step "%s": select action\'s parameters contain an incorrect value.', err.stepName);
        break;

      case CLIENT_ERR.API_INCORRECT_WAIT_ACTION_FIRST_ARGUMENT:
        m('Error on step "%s": wait action\'s first parameter contains an incorrect value.', err.stepName);
        break;
    }
    return msg;
};

exports.build = function(err) {
    if (util.isArray(err)) {
        var msgs = [];
        err.forEach(function(err) {
            msgs.push(generateMsg(err));
        });
        return msgs;
    } else {
        return generateMsg(err);
    }
};