var zlib = require("zlib"), util = require("util"), whacko = require("whacko"), iconv = require("iconv-lite"), sharedConst = require("./../shared/const"), urlUtil = require("./../shared/url_util"), pageProc = require("./../shared/page_processor"), ERR = require("./server_errs");

var ZLIB_INFLATE_INVALID_ENCODING_ERR_CODE = "Z_DATA_ERROR", CHARSET_LIST = [ "iso-8859-1", "iso-8859-2", "iso-8859-3", "iso-8859-4", "iso-8859-5", "iso-8859-6", "iso-8859-7", "iso-8859-8", "iso-8859-9", "iso-8859-10", "iso-8859-11", "iso-8859-12", "iso-8859-13", "iso-8859-14", "iso-8859-15", "iso-8859-16", "windows-1250", "windows-1251", "windows-1252", "windows-1253", "windows-1254", "windows-1255", "windows-1256", "windows-1257", "windows-1258", "windows-874", "windows-866", "koi8-r", "koi8-u", "utf-8", "utf-16", "utf-32", "shift-jis", "x-euc", "big5", "euc-kr" ];

function parseCharset(charset) {
    for (var i = 0; i < CHARSET_LIST.length; i++) {
        var currentCharset = CHARSET_LIST[i], charsetRegExp = new RegExp(currentCharset.replace(/-/g, ""), "i");
        if (charsetRegExp.test(charset.replace(/-/g, ""))) {
            return currentCharset;
        }
    }
    return charset;
}

function concatChunks(chunks) {
    if (chunks.length === 1) {
        return chunks[0];
    }
    var length = 0;
    chunks.forEach(function(chunk) {
        length += chunk.length;
    });
    var data = new Buffer(length), copyPos = 0;
    chunks.forEach(function(chunk) {
        chunk.copy(data, copyPos, 0, chunk.length);
        copyPos += chunk.length;
    });
    return data;
}

function getRawData(chunks, encoding, charset, callback) {
    var encodedData = concatChunks(chunks), decodingCallback = function(err, decodingRes) {
        if (err && !decodingRes) {
            callback({
                code: ERR.INJECTOR_RESOURCE_DECODING_FAILED,
                encoding: encoding
            });
        } else {
            try {
                var rawData = iconv.decode(decodingRes, parseCharset(charset));
            } catch (err) {
                callback({
                    code: ERR.INJECTOR_RESOURCE_CHARSET_DECODING_FAILED,
                    charset: charset
                });
                return;
            }
            callback(null, rawData);
        }
    };
    switch (encoding) {
      case "gzip":
        zlib.gunzip(encodedData, decodingCallback);
        break;

      case "deflate":
        zlib.inflate(encodedData, function(err, decodingRes) {
            if (err && err.code === ZLIB_INFLATE_INVALID_ENCODING_ERR_CODE) {
                zlib.inflateRaw(encodedData, decodingCallback);
            } else {
                decodingCallback(err, decodingRes);
            }
        });
        break;

      default:
        decodingCallback(null, encodedData);
        break;
    }
}

function getEncodedData(rawData, encoding, charset, callback) {
    var data = iconv.encode(rawData, parseCharset(charset));
    switch (encoding) {
      case "gzip":
        zlib.gzip(data, callback);
        break;

      case "deflate":
        zlib.deflate(data, callback);
        break;

      default:
        callback(null, data);
        break;
    }
}

function inject(chunks, encoding, charset, callback, processor) {
    getRawData(chunks, encoding, charset, function(decodingErr, rawData) {
        if (decodingErr) {
            callback(decodingErr);
            return;
        }
        try {
            var processedData = processor(rawData, charset);
            if (processedData === null) {
                return;
            }
        } catch (err) {
            callback(err);
            return;
        }
        getEncodedData(processedData, encoding, charset, function(encodingErr, encodedHtml) {
            if (encodingErr) {
                callback({
                    code: ERR.INJECTOR_RESOURCE_ENCODING_FAILED,
                    encoding: encoding
                });
            } else {
                callback(null, encodedHtml);
            }
        });
    });
}

exports.injectInPage = function(chunks, encoding, charset, injectionOptions, callback) {
    var parseCharset = function($) {
        var needPragma = null, currentCharset = null;
        $("meta").each(function(index, meta) {
            var $meta = $(meta), httpEquivAttr = $meta.attr("http-equiv"), contentAttr = $meta.attr("content"), charsetAttr = $meta.attr("charset");
            if (needPragma !== false && httpEquivAttr && httpEquivAttr.toLowerCase() === "content-type" && contentAttr) {
                var charsetExec = /charset ?= ?['"]?([^ ;"']*)['"]?/.exec(contentAttr);
                if (charsetExec && charsetExec[1]) {
                    needPragma = true;
                    currentCharset = charsetExec[1];
                }
            }
            if (charsetAttr) {
                needPragma = false;
                currentCharset = charsetAttr;
            }
        });
        return currentCharset;
    };
    var rawDataProcessor = function(rawData, actualCharset) {
        try {
            var $ = whacko.load(rawData);
        } catch (parseErrs) {
            throw {
                code: ERR.INJECTOR_DOCUMENT_PARSING_FAILED,
                msg: parseErrs
            };
        }
        if (injectionOptions) {
            var pageCharset = parseCharset($);
            if (pageCharset && pageCharset.toLowerCase() !== actualCharset.toLowerCase()) {
                inject(chunks, encoding, pageCharset, callback, rawDataProcessor);
                return null;
            }
            var $base = $("base"), baseUrl = $base.length ? $base[0].attribs["href"] : "";
            urlUtil.executeWithBaseUrl(baseUrl, function() {
                pageProc.processPage($, injectionOptions.urlReplacer);
            });
            var injection = [];
            if (injectionOptions.styleUrl) {
                injection.push('<link rel="stylesheet" type="text/css" class="');
                injection.push(sharedConst.TEST_CAFE_UI_STYLESHEET_CLASSNAME);
                injection.push('"href = "');
                injection.push(injectionOptions.styleUrl);
                injection.push('">');
            }
            if (injectionOptions.scriptUrl) {
                injection.push('<script type="text/javascript" class="');
                injection.push(sharedConst.TEST_CAFE_SCRIPT_CLASSNAME);
                injection.push('" charset="UTF-8" src="');
                injection.push(injectionOptions.scriptUrl);
                injection.push('">');
                injection.push("</script>");
            }
            if (injection.length) {
                $("head").prepend(injection.join(""));
            }
        }
        return $.html();
    };
    inject(chunks, encoding, charset, callback, rawDataProcessor);
};

exports.injectInStylesheet = function(chunks, encoding, charset, urlReplacer, callback) {
    inject(chunks, encoding, charset, callback, function(rawData) {
        return pageProc.processStylesheet(rawData, urlReplacer);
    });
};

exports.injectInScript = function(chunks, encoding, charset, callback) {
    inject(chunks, encoding, charset, callback, function(rawData) {
        return pageProc.processScript(rawData);
    });
};

exports.injectInManifest = function(chunks, encoding, charset, urlReplacer, callback) {
    inject(chunks, encoding, charset, callback, function(rawData) {
        return pageProc.processManifest(rawData, urlReplacer);
    });
};

exports.injectInJSON = function(chunks, encoding, charset, callback) {
    inject(chunks, encoding, charset, callback, function(rawData) {
        var isJSONP = !pageProc.isJSON(rawData);
        return isJSONP ? pageProc.processScript(rawData, true) : rawData;
    });
};